"""
ldapbase.py - basic LDAP functions
(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

Python compability note:
This module only works with Python 1.6+ since all string parameters
are assumed to be Unicode objects and string methods are used instead
string module.
"""

_version = '0.4.0'

import sys,types,re,urllib,ldap

MOD_STR = {
  ldap.MOD_ADD:'add',
  ldap.MOD_REPLACE:'replace',
  ldap.MOD_DELETE:'delete'
}

SEARCH_SCOPE_STR = ['base','one','sub']

SEARCH_SCOPE = {
  # default for empty search scope string
  '':ldap.SCOPE_BASE,
  # the search scope strings defined in RFC22xx(?)
  'base':ldap.SCOPE_BASE,
  'one':ldap.SCOPE_ONELEVEL,
  'sub':ldap.SCOPE_SUBTREE
}

attrtype_pattern = ur'[\w;.]+(;[\w_-]+)*'
data_pattern = ur'(([^,]|\\,)+|".*?")'
rdn_pattern = attrtype_pattern + ur'[ ]*=[ ]*' + data_pattern
dn_pattern   = rdn_pattern + r'([ ]*,[ ]*' + rdn_pattern + r')*[ ]*'

host_pattern = ur'([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+|[a-zA-Z]+[a-zA-Z0-9-]*(\.[a-zA-Z]+[a-zA-Z0-9-]*)*)+(:[0-9]*)*'

dc_rdn_pattern = ur'(dc|)[ ]*=[ ]*' + data_pattern
dc_dn_pattern   = dc_rdn_pattern + r'([ ]*,[ ]*' + dc_rdn_pattern + r')*[ ]*'

#rdn_regex   = re.compile('^%s$' % rdn_pattern)
dn_regex      = re.compile(u'^%s$' % unicode(dn_pattern))
host_regex = re.compile(host_pattern)

# Some widely used types
StringType = type('')
UnicodeType = type(u'')
ldapScopeType = type(ldap.SCOPE_SUBTREE)

ROOTDSE_ATTRS = [
  'namingContexts',
  'subschemaSubentry',
  'supportedLDAPVersion',
  'subschemaSubentry',
  'supportedControl',
  'supportedExtension',
  'objectclass',
  'supportedSASLMechanisms',
  'dsServiceName',
  'ogSupportedProfile',
  'netscapemdsuffix',
  'dataversion',
  'dsaVersion',
]

def unicode_list(l,charset='utf-8'):
  """
  return list of Unicode objects
  """
  return [
    unicode(i,charset)
    for i in l
  ]

def unicode_entry(e,charset='utf-8'):
  """
  return dictionary of lists of Unicode objects
  """
  result = {}
  for attrtype,valuelist in e.items():
    result[attrtype]=unicode_list(valuelist,charset)
  return result

def dcdn2dnsdomain(dn=''):
  """convert dc-style DN to DNS domain name (see RFC 2377)"""
  dn_components = explode_dn(dn)
  dns_components = []
  for i in xrange(len(dn_components)-1,-1,-1):
    attrtype,value = dn_components[i].split('=',1)
    if attrtype!='dc':
      break
    dns_components.append(value.strip())
  dns_components.reverse()
  return '.'.join(dns_components)


def dnsdomain2dcdn(domain=''):
  """convert DNS domain name to dc-style DN (see RFC 2377)"""
  return ','.join(
    map(
      lambda x:'dc=%s' % x,
      domain.split('.')
    )
  )


def encode_unicode_list(l,charset='utf-8'):
  """
  Encode the list of Unicode objects with given charset
  and return list of encoded strings
  """
  return map(
    lambda i,c=charset:i.encode(c),
    l
  )

def is_ldap_url(s):
  """Fail-safe wrapper function for ldap.is_ldap_url()"""
  if type(s)==UnicodeType:
    s=s.encode('utf-8')
  try:
    string_is_ldap_url = ldap.is_ldap_url(s)
  except TypeError:
    return 0
  else:
    return string_is_ldap_url

def is_dn(s):
  """returns 1 if s is a LDAP DN"""
  if s:
    rm = dn_regex.match(s)
    return rm!=None
  else:
    return 0

def explode_dn(dn,charset='utf-8'):
  """
  Wrapper function for explode_dn() which returns [] for 
  a zero-length DN
  """
  if not dn:
    return []
  if type(dn)==UnicodeType:
    dn = dn.encode(charset)
  dn_list = ldap.explode_dn(dn.strip())
  if dn_list and dn_list!=['']:
    return map(lambda x:unicode(x.strip(),'utf-8'),dn_list)
  else:
    return []


def normalize_dn(dn):
  result = explode_dn(dn)
  return ','.join(result)


def match_dn(dn1,dn2):
  """Return how much levels two dn1 and dn2 are matching (integer)"""
  dn1_list,dn2_list = explode_dn(dn1),explode_dn(dn2)
  if not dn1_list or not dn2_list:
    return (0,u'')
  # Determine dn1_cmp that it's shorter than dn2_cmp
  if len(dn1_list)<=len(dn2_list):
    dn1_cmp,dn2_cmp = dn1_list,dn2_list
  else:
    dn1_cmp,dn2_cmp = dn2_list,dn1_list
  i = 1 ; dn1_len = len(dn1_cmp)
  while (dn1_cmp[-i].lower()==dn2_cmp[-i].lower()):
    i = i+1
    if i>dn1_len:
      break
  if i>1:
    return (i-1,','.join(dn2_cmp[-i+1:]))
  else:
    return (0,u'')


def match_dnlist(dn,dnlist):
  """find best matching parent DN of dn in dnlist"""
  max_match_level, max_match_name = 0, ''
  for dn_item in dnlist:
    match_level, match_name = match_dn(dn_item,dn)
    if match_level>max_match_level:
      max_match_level, max_match_name = match_level, match_name
  return max_match_name

def extract_referrals(e):
  """Extract the referral LDAP URL from a
     ldap.PARTIAL_RESULTS exception object"""
  if e.args[0].has_key('info'):
    info, ldap_url_info = map(
      lambda x:x.strip(),
      e.args[0]['info'].split('\n',1)
    )
  else:
    raise ValueError, "Referral exception object does not have info field"
  ldap_urls = map(lambda x:x.strip(),ldap_url_info.split('\n'))
  matched = e.args[0].get('matched',None)
  return (matched,ldap_urls)

def ParentDN(dn):
  """returns parent-DN of dn"""
  dn_comp = explode_dn(dn)
  if len(dn_comp)>1:
    return ','.join(dn_comp[1:])
  elif len(dn_comp)==1:
    return ''
  else:
    return None

def SplitRDN(dn):
  """returns tuple (RDN,base DN) of dn"""
  dn_comp = explode_dn(dn)
  if len(dn_comp)>1:
    return dn_comp[0], ','.join(dn_comp[1:])
  elif len(dn_comp)==1:
    return dn_comp[0], ''
  else:
    return None

def ParentDNList(dn,rootdn=''):
  """returns a list of parent-DNs of dn"""
  result = []
  DNComponentList = explode_dn(dn)
  if rootdn:
    max_level = len(explode_dn(rootdn))
  else:
    max_level = len(DNComponentList)
  for i in range(1,max_level):
    result.append(','.join(DNComponentList[i:]))
  return result


class BAD_LDAPURL(ldap.LDAPError):
  """
  Simple exception class raised when a LDAP URL
  contains wrong values
  """
  pass

def parse_ldap_url(ldap_url):
  """
  parse a LDAP URL and return (host,dn,attrs,scope,filter)

  host         LDAP host
  dn           distinguished name (search root)
  attrs        list of attribute types
  scope        integer search scope for ldap-module
  filter       LDAP search filter
  extensions   list of extensions
  """
  dn = ''
  attr_list = None
  search_scope = ldap.SCOPE_BASE
  ldap_filter = '(objectclass=*)'
  extensions = []
  dummy,rest = ldap_url.split('://',1)
  try:
    host,rest = rest.split('/',1)
  except ValueError:
    raise BAD_LDAPURL, 'LDAP URL must contain third slash.'
  else:
    paramlist=rest.split('?')
    paramlist_len = len(paramlist)
    if paramlist_len>=1:
      dn = urllib.unquote_plus(paramlist[0])
    if (paramlist_len>=2) and (paramlist[1]):
      attr_list = urllib.unquote_plus(paramlist[1].strip()).split(',')
    if paramlist_len>=3:
      try:
        search_scope = SEARCH_SCOPE[paramlist[2].strip()]
      except KeyError:
        raise BAD_LDAPURL, "Search scope must be either one of base, one or sub. LDAP URL contained %s" % repr(paramlist[2])
    if paramlist_len>=4:
      ldap_filter = urllib.unquote_plus(paramlist[3].strip())
    if paramlist_len>=5:
      extensions = paramlist[4].strip().split(',')
  return (host,dn,attr_list,search_scope,ldap_filter,extensions)


def create_ldap_url(
  hostport='',		# host:port of LDAP server
  dn='',		# DN of search root
  attrs=[],		# list with attribute types
  scope=ldap.SCOPE_BASE,# Search scope
  			# (can be either string or constant from ldap-module)
  filterstr='(objectclass=*)',# LDAP filter string according to RFC2254
  extensions='',	# Extensions
  urlencode=0,		# Apply URL-encoding
  charset='utf-8'       # Charset for encoding Unicode objects
):
  """
  build LDAP URL of search query according to RFC2255

  hostport     host:port
  dn           distinguished name
  attributes   list with attributes
  scope        search scope string
  filter       LDAP search filter
  ldapurl    = scheme "://" [hostport] ["/"
                   [dn ["?" [attrs] ["?" [scope]
                   ["?" [filter] ["?" extensions]]]]]]
  """
  def urlencoding(s):
    return urllib.quote(s).replace(',','%2C').replace('/','%2F')
  if attrs:
    attrs_str = ','.join(attrs)
  else:
    attrs_str = ''
  if type(scope)==ldapScopeType:
    scope_str = SEARCH_SCOPE_STR[scope]
  else:
    scope_str = scope
  v = vars()
  v['hostport'] = hostport.encode('ascii')
  for var in ['dn','filterstr','attrs_str','scope_str']:
    if type(v[var])==UnicodeType:
      v[var]=v[var].encode(charset)
    if urlencode:
      v[var]=urlencoding(v[var])
  return 'ldap://%(hostport)s/%(dn)s?%(attrs_str)s?%(scope_str)s?%(filterstr)s' % v


def AttrEncode(attrtype,value,binary_attrtypes=[],charset='utf-8'):
  if attrtype in binary_attrtypes:
    return str(value)
  else:
    return value.encode(charset)


def add_modifylist(entry,ldap_charset='utf-8',binary_attrtypes=[]):
  """Build modify list for call of method add()"""
  modlist = []
  for attrtype in entry.keys():
    data = [
      AttrEncode(attrtype.lower(),d,binary_attrtypes,ldap_charset)
      for d in entry[attrtype]
    ]
    modlist.append((attrtype.encode('ascii'),data))
  return modlist


def modify_modifylist(
  old_entry,
  new_entry,
  ldap_charset='utf-8',
  binary_attrtypes=[]
):
  """Build differential modify list for call of method modify()"""
  modlist = []
  old_entry_keys = map(lambda x:x.lower(),old_entry.keys())
  for attrtype in new_entry.keys():
    attrtype_lower = attrtype.lower()
    if attrtype_lower in old_entry_keys:
      old_data = unicode_list(
        old_entry.get(attrtype,old_entry.get(attrtype_lower,[])),
        ldap_charset
      )
      if new_entry[attrtype]!=old_data:
	if old_data and (new_entry[attrtype]==[]) or (new_entry[attrtype]==['']):
	  # delete an existing attribute
	  modlist.append(
            (
              ldap.MOD_DELETE,
              attrtype.encode(ldap_charset),
              map(lambda d,cs=ldap_charset:d.encode(cs),old_data)
            )
          )
	else:
	  # modify an existing attribute
	  modlist.append(
            (
              ldap.MOD_REPLACE,
              attrtype.encode('ascii'),
              map(lambda d,cs=ldap_charset:d.encode(cs),new_entry[attrtype])
            )
          )
    else:
      for d in new_entry[attrtype]:
	if d:
          data = AttrEncode(attrtype.lower(),d,binary_attrtypes,ldap_charset)
	  # append new attribute
          modlist.append((ldap.MOD_ADD,attrtype.encode('ascii'),data))
  return modlist

def escape_filter_chars(s):
  """RFC2254: Convert special characters in search filter to escaped hex"""
  for ch in ['\\','*','(',')','\000']:
    s = s.replace(ch,'\%02X' % ord(ch))
  return s

def escape_binary_dnchars(s):
  """Convert NON-ASCII characters in DN to escaped hex form"""
  new = ''
  for ch in s:
    c=ord(ch)
    if (c<32) or (c>=128):
      new = new+('\%02X' % c)
    else:
      new = new+ch
  return new


# Regex object used for finding hex-escaped characters
hex_dnchars_regex = re.compile(r'\\[0-9a-fA-F][0-9a-fA-F]')

def unescape_binary_dnchars(s,charset='utf-8'):
  """Convert hex-escaped characters in DN to binary NON-ASCII chars"""
  if type(s)==UnicodeType:
    s = s.encode(charset)
  asciiparts = hex_dnchars_regex.split(s)
  hexparts = hex_dnchars_regex.findall(s)
  new = asciiparts[0]
  for i in range(1,len(asciiparts)):
    new = chr(eval('0x%s' % hexparts[i-1][1:])).join([new,asciiparts[i]])
  return new


def SearchTree(l=None,ldap_dn='',attrsonly=1,timeout=-1,charset='utf-8'):
  """
  Returns all DNs of the sub tree at ldap_dn as list of
  Unicode string objects.
  """
  if type(ldap_dn)==StringType:
    ldap_dn = unicode(ldap_dn,charset)
  stack, result = [ldap_dn],[]
  while stack:
    dn = stack.pop()
    result.append(dn)
    try:
      r = l.search_st(dn.encode(charset),ldap.SCOPE_ONELEVEL,'(objectclass=*)',['objectclass'],attrsonly,timeout)
    except ldap.NO_SUCH_OBJECT:
      r = []
    stack.extend(
      [
        unicode(dn,charset)
        for dn,entry in r
      ]
    )
  return result


class USERNAME_NOT_FOUND(ldap.LDAPError):
  """
  Simple exception class raised when SmartLogin() does not
  find any entry matching search
  """
  pass

class USERNAME_NOT_UNIQUE(ldap.LDAPError):
  """
  Simple exception class raised when SmartLogin() does not
  find more than one entry matching search
  """
  pass

def SmartLogin(
  l,                        # LDAP connection object created with ldap.open()
  username='',              # User name or complete bind DN (UTF-8 encoded)
  searchroot='',            # search root for user entry search
  filtertemplate='(uid=%s)',# template string for LDAP filter
  scope=ldap.SCOPE_SUBTREE, # search scope
  attrnamesonly=1           # retrieve only attribute names when searching
):
  """
  Do a smart login.
  
  If username is a valid DN it's used as bind-DN without further action.
  Otherwise filtertemplate is used to construct a LDAP search filter
  containing username instead of %s.
  The calling application has to handle all possible exceptions:
  ldap.NO_SUCH_OBJECT, ldap.FILTER_ERROR, ldapbase.USERNAME_NOT_UNIQUE
  ldap.INVALID_CREDENTIALS, ldap.INAPPROPRIATE_AUTH
  """
  if not username:
    return ''
  elif is_dn(username):
    return normalize_dn(username)
  else:
    searchfilter = filtertemplate.replace('%s',username)
    if searchroot is None:
      searchroot = ''
    # Try to find a unique entry with filtertemplate
    try:
      result = l.search_s(searchroot,scope,searchfilter,['objectclass'],attrnamesonly)
    except ldap.NO_SUCH_OBJECT:
      raise USERNAME_NOT_FOUND({'desc':'Smart login did not find a matching user entry.'})
    else:
      if not result:
        raise USERNAME_NOT_FOUND({'desc':'No matching user entry'})
      elif len(result)!=1:
        raise USERNAME_NOT_UNIQUE({'desc':'More than one matching user entries'})
      else:
        return normalize_dn(result[0][0])


def test():
  """Test functions"""

  print '\nTesting function is_dn():'
  ldap_dns = {
    u'o=Michaels':1,
    u'iiii':0
  }
  for ldap_dn in ldap_dns.keys():
    result_is_dn = is_dn(ldap_dn)
    if result_is_dn !=ldap_dns[ldap_dn]:
      print 'is_dn("%s") returns %d instead of %d.' % (
        ldap_dn,result_is_ldap_dn,ldap_dns[ldap_dn]
      )

  print '\nTesting function is_ldap_url():'
  is_ldap_url_tests = {
    # Examples from RFC2255
    u'ldap:///o=University%20of%20Michigan,c=US':1,
    u'ldap://ldap.itd.umich.edu/o=University%20of%20Michigan,c=US':1,
    u'ldap://ldap.itd.umich.edu/o=University%20of%20Michigan,':1,
    u'ldap://host.com:6666/o=University%20of%20Michigan,':1,
    u'ldap://ldap.itd.umich.edu/c=GB?objectClass?one':1,
    u'ldap://ldap.question.com/o=Question%3f,c=US?mail':1,
    u'ldap://ldap.netscape.com/o=Babsco,c=US??(int=%5c00%5c00%5c00%5c04)':1,
    u'ldap:///??sub??bindname=cn=Manager%2co=Foo':1,
    u'ldap:///??sub??!bindname=cn=Manager%2co=Foo':1,
    # More examples from various sources
    u'ldap://ldap.nameflow.net:1389/c%3dDE':1,
    u'ldap://root.openldap.org/dc=openldap,dc=org':1,
    u'ldap://root.openldap.org/dc=openldap,dc=org':1,
    u'ldap://x500.mh.se/o=Mitthogskolan,c=se????1.2.752.58.10.2=T.61':1,
    u'ldp://root.openldap.org/dc=openldap,dc=org':0,
  }
  for ldap_url in is_ldap_url_tests.keys():
    result_is_ldap_url = is_ldap_url(ldap_url)
    if result_is_ldap_url !=is_ldap_url_tests[ldap_url]:
      print 'is_ldap_url("%s") returns %d instead of %d.' % (
        repr(ldap_url),result_is_ldap_url,is_ldap_url_tests[ldap_url]
      )

  print '\nTesting function parse_ldap_url():'
  parse_ldap_url_tests = {
    u'ldap://root.openldap.org/dc=openldap,dc=org':(
      u'root.openldap.org', u'dc=openldap,dc=org', [], 0, '(objectclass=*)', []
    ),
    u'ldap://x500.mh.se/o=Mitthogskolan,c=se????1.2.752.58.10.2=T.61':(
      u'x500.mh.se', u'o=Mitthogskolan,c=se', [], 0, u'', [u'1.2.752.58.10.2=T.61']
    ),
  }
  for ldap_url in parse_ldap_url_tests.keys():
    result_parse_ldap_url = parse_ldap_url(ldap_url)
    if result_parse_ldap_url !=parse_ldap_url_tests[ldap_url]:
      print 'parse_ldap_url(%s) returns:\n%s\ninstead of:\n%s\n' % (
        repr(ldap_url),
        repr(result_parse_ldap_url),
        repr(parse_ldap_url_tests[ldap_url])
      )

  print '\nTesting functions escape_binary_dnchars() and unescape_binary_dnchars():'
  binary_strings = [
    '\000\001\002\003\004\005\006',
    'Michael Strder',
  ]
  for s in binary_strings:
    if unescape_binary_dnchars(escape_binary_dnchars(s))!=s:
      print 'Testing of unescape_binary_dnchars(escape_binary_dnchars()) failed for string %s.' % (
        repr(s)
      )

  print '\nTesting function match_dn():'
  match_dn_tests = {
    ('O=MICHAELS','o=michaels'):(1,u'O=MICHAELS'),
    ('CN=MICHAEL STROEDER,O=MICHAELS','o=michaels'):(1,u'O=MICHAELS'),
    ('CN=MICHAEL STROEDER,O=MICHAELS',''):(0,u''),
    ('CN=MICHAEL STROEDER,O=MICHAELS','     '):(0,u''),
    ('CN=MICHAEL STROEDER,O=MICHAELS','  cn=Michael Stroeder,o=Michaels  '):(2,u'cn=Michael Stroeder,o=Michaels'),
    ('CN=MICHAEL STROEDER,O=MICHAELS','mail=michael@stroeder.com,  cn=Michael Stroeder,o=Michaels  '):(2,u'cn=Michael Stroeder,o=Michaels'),
  }
  for dn1,dn2 in match_dn_tests.keys():
    result_match_dn = match_dn(dn1,dn2)
    if result_match_dn[0] !=match_dn_tests[(dn1,dn2)][0] or \
       result_match_dn[1].lower() !=match_dn_tests[(dn1,dn2)][1].lower():
      print 'match_dn(%s,%s) returns:\n%s\ninstead of:\n%s\n' % (
        repr(dn1),repr(dn2),
        repr(result_match_dn),
        repr(match_dn_tests[(dn1,dn2)])
      )


if __name__ == '__main__':
  test()
