"""
ldapsession.py - class library for LDAP sessions
(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)
"""

_version = '0.0.9'

import socket, time, threading, ldap, ldapbase

from ldapthreading import open as ldap_open
#from ldap import open as ldap_open

# Constants for options
OPT_OFF=0
OPT_ON=1
OPT_DICT={'referrals':ldap.OPT_REFERRALS,'restart':ldap.OPT_RESTART}

OPTIONS_ATTRS = ['referrals','restart','deref']

# Used attributes from RootDSE
ROOTDSE_ATTRS = [
  'namingContexts',
  'subschemaSubentry',
  'supportedLDAPVersion',
  'subschemaSubentry',
]

# LDAPSession object attributes which are pickled.
LDAPSESSION_ATTRS = [
  'host','hostAddress','who','cred','auth_method','dn','timeout',
  'openldap2','currentLDAPVersion','charset',
  'read_cache',
]+ROOTDSE_ATTRS+OPTIONS_ATTRS

READ_CACHE_EXPIRE = 20

class LDAPSession:

  def __init__(self):
    """Initilialize a LDAPSession object"""
    self.searchRoot = u''
    self.namingContexts = []
    self.dn = u''
    # Authentication method used.
    self.auth_method = ldap.AUTH_SIMPLE
    # Character set/encoding of data stored on this particular host
    self.charset = 'utf-8'
    # This is a dictionary for storing arbitrary objects
    # tied to a LDAP session
    self.read_cache = {}

  def _open(self,host,max_conn_try=1,conn_try_delay=0.050):
    retry_counter = 0
    while 1:
      retry_counter += 1
      try:
        l = ldap_open(host)
      except ldap.LDAPError,e:
        if retry_counter>=max_conn_try:
          raise e
        time.sleep(conn_try_delay)
      else:
        l.set_rebind_proc(None)
        break
    return l

  def open(self,host=None,timeout=0):
    """Open a LDAP connection with separate DNS lookup"""
    self.timeout = timeout
    # Do DNS lookup here
    h = host.encode('ascii').split(':')
    if len(h)>=2:
      hp = ':'.join([socket.gethostbyname(h[0]),h[1]])
    else:
      hp = socket.gethostbyname(h[0])
    # Try connecting to LDAP host
    self.l = self._open(hp)
    self.host = host
    # Store IP address:port for faster reconnect
    self.hostAddress = hp
    self.openldap2=hasattr(self.l,'version')

  def unbind(self):
    """Close LDAP connection object if necessary"""
    if hasattr(self,'l'):
      try:
        self.l.unbind_s()
      except ldap.LDAPError:
        pass
      except AttributeError:
        pass
      del self.l

  def getUmichConfig(self):
    """
    Try to read entry cn=config attribute database
    from UMich LDAPv2 server derivate
    """
    try:
      ldap_result = self.readEntry(
        'cn=config',['database'],
      )
    except ldap.NO_SUCH_OBJECT:
      result = []
    except ldap.PARTIAL_RESULTS:
      result = []
    except ldap.UNDEFINED_TYPE:
      result = []
    except ldap.INAPPROPRIATE_MATCHING:
      result = []
    except ldap.OPERATIONS_ERROR:
      result = []
    else:
      if ldap_result:
	result = []
	l = ldap_result[0][1].get('database',[])
	for d in l:
          try:
            dbtype,basedn = d.split(' : ',1)
	  except ValueError:
	    pass
	  else:
            result.extend(map(lambda x:x.strip(),basedn.split(' : ')))
	result = result
      else:
	result = []
    return [
      unicode(i,self.charset)
      for i in result
    ]

  def getRootDSE(self):
    """retrieve attributes from Root DSE"""
    if not hasattr(self,'who'):
      self.bind('','',ldap.AUTH_SIMPLE)
    try:
      ldap_result = self.readEntry(
        '',ROOTDSE_ATTRS
      )
    except ldap.NO_SUCH_OBJECT:
      pass
    except ldap.PARTIAL_RESULTS:
      pass
    except ldap.UNDEFINED_TYPE:
      pass
    except ldap.INAPPROPRIATE_MATCHING:
      pass
    except ldap.OPERATIONS_ERROR:
      pass
    except ldap.UNWILLING_TO_PERFORM:
      pass
    else:
      if ldap_result:
        rootDSE=ldap_result[0][1]
      else:
        rootDSE={}
      for attrtype in ROOTDSE_ATTRS:
        if rootDSE.has_key(attrtype):
          value = ldapbase.unicode_list(rootDSE[attrtype],self.charset)
          setattr(self,attrtype,value)
    if not hasattr(self,'namingContexts') or not self.namingContexts:
      self.namingContexts=self.getUmichConfig()
    self.namingContexts = map(ldapbase.normalize_dn,self.namingContexts)
    if self.namingContexts:
      self.searchRoot=ldapbase.match_dnlist(self.dn,self.namingContexts)

  def getObjectClasses(self,dn):
    """Get a list of object classes associated with an entry"""
    entry = self.readEntry(dn,['objectClass'])[0][1]
    return entry.get('objectClass',entry.get('objectclass',[]))

  def getAttributeTypes(self,dn,onlyattrtypes=1):
    """Get a list of object classes associated with an entry"""
    if dn:
      attrs = None
    else:
      attrs = ['+']+ldapbase.ROOTDSE_ATTRS
    search_result = self.readEntry(
      dn,attrs,onlyattrtypes
    )
    if search_result:
      return search_result[0][1].keys()
    else:
      raise ldap.NO_SUCH_OBJECT

  def readEntry(self,dn,attrtype_list=None,only_attrtypes=0):
    """Read a single entry"""
    read_cache_id = ldapbase.create_ldap_url(self.host,dn,attrtype_list)
    if self.read_cache.has_key(read_cache_id):
      timestamp, read_cache_result = self.read_cache[read_cache_id]
      if timestamp+READ_CACHE_EXPIRE>time.time():
        return read_cache_result
      else:
        del self.read_cache[read_cache_id]
    # Read single entry
    search_result = self.l.search_st(
      dn.encode(self.charset),
      ldap.SCOPE_BASE,
      "(objectclass=*)",
      attrtype_list,
      0,
      self.timeout
    )
    if search_result:
      # Put the read entry in the time-stamped read_cache
      self.read_cache[read_cache_id] = (time.time(),search_result)
    return search_result

  def existingEntry(self,dn,suppress_referrals=0):
    """Returns 1 if entry exists, 0 if NO_SUCH_OBJECT was raised."""
    try:
      self.readEntry(dn,[])
    except ldap.NO_SUCH_OBJECT:
      return 0
    except ldap.PARTIAL_RESULTS,e:
      if suppress_referrals:
        return 0
      else:
        raise e
    else:
      return 1
      
  def addEntry(self,dn,modlist):
    """Add single entry"""
    self.l.add_s(dn.encode(self.charset),modlist)
    return

  def modifyEntry(self,dn,modlist):
    """Modify single entry"""
    self.l.modify_s(dn.encode(self.charset),modlist)
    # Deleted cached entry data
    self.l.uncache_entry(dn.encode(self.charset))
    read_cache_id = ldapbase.create_ldap_url(self.host,dn,[])
    if self.read_cache.has_key(read_cache_id):
      del self.read_cache[read_cache_id]
    return

  def modRDN(self,dn,new_rdn,delold):
    """Modify single entry"""
    self.l.modrdn_s(
      dn.encode(self.charset),
      new_rdn.encode(self.charset),
      delold
    )
    # Deleted cached entry data
    self.l.uncache_entry(dn.encode(self.charset))
    read_cache_id = ldapbase.create_ldap_url(self.host,dn,[])
    if self.read_cache.has_key(read_cache_id):
      del self.read_cache[read_cache_id]
    return

  def deleteEntry(self,dn):
    """Delete single entry"""
    self.l.delete_s(dn.encode(self.charset))
    # Deleted cached entry data
    self.l.uncache_entry(dn.encode(self.charset))
    read_cache_id = ldapbase.create_ldap_url(self.host,dn,[])
    if self.read_cache.has_key(read_cache_id):
      del self.read_cache[read_cache_id]
    return

  def setDN(self,dn):
    """Set currently used DN"""
    dn=ldapbase.normalize_dn(dn)
    if dn!=self.dn:
      self.dn = dn
      self.searchRoot=ldapbase.match_dnlist(self.dn,self.namingContexts)

  def set_option(self,name,value):
    """Set LDAPObject option or attribute"""
    if self.openldap2 or name=='deref':
      setattr(self.l,name,value)
    else:
      if name in ['referrals','restart']:
        value = value<<OPT_DICT[name]
        if value==OPT_ON:
          self.l.options = self.l.options | (value << OPT_DICT[name])
        elif value==OPT_OFF:
          self.l.options = self.l.options & (~value <<OPT_DICT[name])
        else:
          raise ValueError,'Argument value must be either one of ldapsession.OPT_ON or ldapsession.OPT_OFF.'
      else:
        raise AttributeError,'Wrong option name %s.' % (name)
    setattr(self,name,value)

  def bind(
    self,
    who='',
    cred='',
    method=ldap.AUTH_SIMPLE,
    filtertemplate='(uid=%s)',
    attrnamesonly=1
  ):
    """
    Bind to host
    
    With OpenLDAP 2.0 this methods tries to bind via LDAPv3 and falls back to
    LDAPv2 if the server returns an INVALID_PROTOCOL error.
    """
    # Smart login
    if who and not ldapbase.is_dn(who):
      if self.searchRoot is None:
        self.getRootDSE()
      who = ldapbase.SmartLogin(
        self.l,
        who,
        self.searchRoot,
        filtertemplate,
        attrnamesonly,
      )
    # Bind
    if self.openldap2:
      self.l.version=3
      try:
        self.l.bind_s(who.encode(self.charset),cred.encode(self.charset),method)
      except ldap.PROTOCOL_ERROR,e:
        if hasattr(e,'args') and \
           type(e.args)==type(()) and \
           type(e.args[0])==type({}) and \
           e.args[0].get('info','').lower()=='version not supported':
          self.l.version=2
        else:
          raise e
      else:
        self.who = who #; self.cred=cred ; self.auth_method = method
        self.currentLDAPVersion=self.l.version
        return
    self.l.bind_s(who.encode(self.charset),cred.encode(self.charset),method)
    self.who = who
    self.currentLDAPVersion = 2
    return

  def setCache(self,timeout=0,maxmem=0):
    """
    Enable or disable caching in python-ldap's LDAPObject
    
    If timeout is zero or maxmem is zero the caching is completely
    switched off by calling self.l.destroy_cache().
    """
    if not maxmem or not timeout and hasattr(self.l,'destroy_cache'):
      self.l.destroy_cache()
    elif hasattr(self.l,'enable_cache'):
      self.l.enable_cache(timeout,maxmem)

  def valid(self):
    """return 1 if connection is valid"""
    return hasattr(self,'l') and self.l.valid

  def __getstate__(self):
    """return data representation for pickled LDAPSession object"""
    d = {}
    for attr in LDAPSESSION_ATTRS:
      if hasattr(self,attr):
        d[attr]=getattr(self,attr)
    return d

  def __setstate__(self,d):
    """restore LDAPSession object"""
    for attr in d.keys():
      setattr(self,attr,d[attr])
    if hasattr(self,'hostAddress'):
      self.l = self._open(self.hostAddress)
      for attr in OPTIONS_ATTRS:
        if hasattr(self,attr):
          self.set_option(attr,getattr(self,attr))
      if self.openldap2:
        self.l.version=self.currentLDAPVersion
      if hasattr(self,'who'):
        self.l.bind_s(self.who.encode(self.charset),self.cred.encode(self.charset),self.auth_method)
    if hasattr(self,'dn') and hasattr(self,'namingContexts'):
      self.searchRoot=ldapbase.match_dnlist(self.dn,self.namingContexts)

  def __repr__(self):
    return '<LDAPSession%s: %s>' % (
      ' connected'*self.valid(),
      ','.join(
        map(
          lambda a,s=self:'%s:%s' % (a,repr(getattr(s,a))),
          filter(lambda a,s=self:hasattr(s,a),
            [
              'host','who','dn',
              'currentLDAPVersion',
            ]
          )
        )
      )
    )
