"""
w2lcore.py: some core functions used throughout web2ldap

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>
Distributed under GNU Public License (GPL)
"""

Version = '0.9.0rc2'

import sys,os,types,time,re,threading,traceback,\
       cgiforms,cgisession,msbase,ldap,ldapbase,cgissl, \
       web2ldapcnf,w2lcnf,msgzip,ldapoc

from ldapbase import ParentDN,ParentDNList,is_dn,normalize_dn,dn_pattern

startUpTime = time.time()

def utf2display(dispcharset,data):
  if type(data)==types.UnicodeType:
    return cgiforms.escapeHTML(data.encode(dispcharset))
  elif type(data) is types.StringType:
    if data[0:6]=='{T.61}':
      return cgiforms.escapeHTML(unicode(data[6:],'iso-8859-1').encode(dispcharset))
    else:
      try:
        return cgiforms.escapeHTML(unicode(data,'utf-8').encode(dispcharset))
      except:
        return cgiforms.escapeHTML(unicode(data,'iso-8859-1').encode(dispcharset))
  elif type(data) is types.ListType:
    return [
      utf2display(dispcharset,i)
      for i in data
    ]
  elif data is None:
    return ''
  else:
    raise TypeError, "expected string or list"

def iso2display(dispcharset,data):
  if type(data) is types.StringType:
    return cgiforms.escapeHTML(unicode(data,'iso-8859-1').encode(dispcharset))
  elif type(data) is types.ListType:
    l = []
    for i in data:
      l.append(utf2display(dispcharset,i))
    return l
  elif data is None:
    return ''
  else:
    raise TypeError, "expected string or list"

def input2utf(inputcharset,s):
  if type(s)==types.UnicodeType:
    return s.encode('utf-8')
  else:
    return unicode(s,inputcharset).encode('utf-8')

########################################################################
# Initialize some constants
########################################################################

## Instantiate web session object
session = cgisession.CGISession(
  expire_relogin=web2ldapcnf.session_relogin,
  expire_remove=web2ldapcnf.session_remove,
  crosscheckvars = web2ldapcnf.session_checkvars,
)

class CleanUpThread(threading.Thread):
  """
  Thread class for clean-up thread
  """
  def __init__(self,cgiSession,interval,session_expire,session_remove):
    self.enabled = 1
    self._removed = 0
    v = vars()
    for attr in ['cgiSession','interval','session_expire','session_remove']:
      setattr(self,attr,v[attr])
    threading.Thread.__init__(self,name=self.__class__.__module__+self.__class__.__name__)

  def __repr__(self):
    return '%s: %d sessions removed' % (
      self.getName(),self._removed
    )

  def run(self):
    """Thread function for cleaning up session database"""
    while self.enabled:
      try:
        current_time = time.time()
        sessiondict_keys = filter(
          lambda id:not (id is None or id.startswith('__')),
          self.cgiSession.sessiondict.keys()
        )
        for session_id in sessiondict_keys:
          try:
            session_timestamp,ldapSession = self.cgiSession.sessiondict[session_id]
          except KeyError:
            # Avoid race condition. The session might have been
            # deleted in the meantime. But make sure everything is deleted.
            self.cgiSession.deleteSession(session_id)
          else:
            # Check expiration time
            if (session_timestamp+self.session_expire<current_time):
              if ldapSession=='_created_':
                # Completely hanging sessions are removed immediately
                self.cgiSession.deleteSession(session_id)
                self._removed+=1
              else:
                # Disconnect from LDAP server
                ldapSession.unbind()
            if session_timestamp+self.session_remove<current_time:
              # Remove expired session
              self.cgiSession.deleteSession(session_id)
              self._removed+=1
      except:
        # Catch all exceptions to avoid thread being killed.
        if __debug__:
          traceback.print_exc()
        pass

      # Sleeping until next turn
      time.sleep(self.interval)

    return # CleanUpThread.run()


# Mapping from sndhdr.what() result to MIME-type
sndhdr_mimetype = {
  'au':'audio/basic',
  'wav':'audio/wav',
  'aiff':'audio/x-aiff',
  'aifc':'audio/x-aiff',
}
# Mapping from imghdr.what() result to MIME-type
imghdr_mimetype = {
  'bmp':'image/bmp',
  'gif':'image/gif',
  'jpeg':'image/jpeg',
  'pbm':'image/x-portable-bitmap',
  'pgm':'image/x-portable-graymap',
  'png':'image/png',
  'ppm':'image/x-portable-pixmap',
  'rast':'image/x-cmu-raster',
  'rgb':'image/x-rgb',
  'tiff':'image/tiff',
  'xbm':'image/x-xbitmap',
}

ldap_binaryattr = msbase.CaseinsensitiveStringKeyDict(web2ldapcnf.ldap_binaryattr)
ldap_binaryattrkeys = map(lambda i,u=unicode:u(i,'iso-8859-1'),ldap_binaryattr.keys())

# Define a attribute name <-> user friendly attribute name mapping
ldap_knownattr = msbase.CaseinsensitiveStringKeyDict(web2ldapcnf.ldap_knownattr)
for i in ldap_binaryattrkeys:
  ldap_knownattr[unicode(i)] = ldap_binaryattr[i][0]

ldap_known_objectclasses = ldapoc.ldap_oc.keys()
ldap_known_objectclasses.sort()

search_options = (
  (ur'(%s=*%s*)',u'contains'),
  (ur'(!(%s=*%s*))',u"doesn't contain"),
  (ur'(%s=%s)',u'is'),
  (ur'(!(%s=%s))',u'is not'),
  (ur'(%s=%s*)',u'begins with'),
  (ur'(%s=*%s)',u'ends with'),
  (ur'(%s~=%s)',u'sounds like'),
  (ur'(%s>=%s)',u'greater equal than'),
  (ur'(%s<=%s)',u'lesser equal than')
)

class ErrorExitClass:
  """Base class for web2ldap application exceptions"""
  def __init__(self,ls,Msg):
    self.ls = ls
    self.Msg = Msg

def log_exception(errf,sid,command,form,ls,env):
  """
  Write an exception with environment vars, LDAP connection data
  and Python traceback to errf file object.
  """
  logentry = ['%s\nweb2ldap error at %s' % (
      '-'*60,
      time.strftime(
        '%Y-%m-%dT%H:%M:%SZ',time.gmtime(time.time())
      ),
    )]
  logentry.append('LDAPSession instance: %s' % repr(ls))
  env_vars = env.keys();env_vars.sort()
  for v in env_vars:
    logentry.append('%s: "%s"' % (v,env[v]))
  exc_obj,exc_value,exc_traceback = sys.exc_info()
  logentry.append(''.join(traceback.format_exception(exc_obj,exc_value,exc_traceback,20)))
  # Avoid memory leaks
  exc_obj=None;exc_value=None;exc_traceback=None
  del exc_obj;del exc_value;del exc_traceback
  errf.write(os.linesep.join(logentry))
