"""
w2lgui.py: basic functions for GUI elements

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>
Distributed under GNU Public License (GPL)
"""

import sys, os, types, time, re, urllib, \
       cgiforms, utctime, httphelper, ldap, msbase, ldapbase, \
       web2ldapcnf, msgzip, w2lcore, w2lcnf

from ldapbase import ParentDN,ParentDNList,is_dn,normalize_dn,dn_pattern
from w2lviewer import snd_mimetype

########################################################################
# Initialize some constants used throughout web2ldap
########################################################################

host_pattern = '([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+|[a-zA-Z]+[a-zA-Z0-9-]*(\.[a-zA-Z]+[a-zA-Z0-9-]*)*)?(:[0-9]*|)?'
url_pattern  = r'^(ftp|http|https|news|snews|ldap|ldaps):(|//)[^ ]*'
url_regex  = re.compile(url_pattern)
labeleduri_regex = re.compile(url_pattern+r' .*')
timestamp_pattern = r'^([0-9]){12,14}Z$'
timestamp_regex  = re.compile(timestamp_pattern)
mail_pattern = r'^[\w@.=/_ -]+@[a-zA-Z0-9-]+(\.[a-zA-Z0-9-]+)*$'
mail_regex = re.compile(mail_pattern)
ldap_binaryattrkeys = web2ldapcnf.ldap_binaryattr.keys()

ROOTDSE_ATTRS_LOWER = map(lambda i:i.lower(),ldapbase.ROOTDSE_ATTRS)

HIDDEN_FIELD = '<input type="hidden" name="%s" value="%s">'

# This function searches for variants
def GetVariantFilename(pathname,variantlist):
  for v in variantlist:
    variant_filename = pathname+'.'+v
    if os.path.isfile(variant_filename):
      break
  else:
    variant_filename = pathname
    if not os.path.isfile(variant_filename):
      variant_filename = ''
  return variant_filename


# Build the string of a button
def CommandButton(
  form,
  command,
  submitstr,
  sid,
  who=None,
  cred=None,
  host=None,
  dn=None,
  extrastr='',
  target='_top',
  hidden_fields = []
):
  charset = form.accept_charset
  form_str = ["""
<form
  id="CommandForm"
  action="%s/%s"
  method="%s"
  target="%s"
  enctype="application/x-www-form-urlencoded"
  accept-charset="%s"
>
"""  % (
    form.script_name,
    command,
    web2ldapcnf.script_method,
    target,
    charset,
  )
]
  # Add optional hidden fields if value is not None
  for param in ['sid','who','cred','host','dn']:
    if vars()[param]!=None:
      form_str.append(HIDDEN_FIELD % (
          param,
	  w2lcore.utf2display(charset,vars()[param])
	)
      )
  for param,paramvalue in hidden_fields:
    form_str.append(HIDDEN_FIELD % (
          param,
	  cgiforms.escapeHTML(paramvalue.encode(charset))
	)
      )
  form_str.append("""
<input id="CommandButton" type="submit" value="%s">
%s
</form>""" % (submitstr,extrastr)
  )
  return '\n'.join(form_str)


def DisplayDN(sid,form,ls,dn,commandbutton=0):
  """Display a DN as LDAP URL with or without button"""
  if dn:
    dn_str = w2lcore.utf2display(form.accept_charset,dn)
  else:
    dn_str = '- World -'
  href_str = '<a href="%s" target="_ldapurl">%s</a>' % (
    ldapbase.create_ldap_url(
      ls.host.encode('ascii'),
      dn,
      urlencode=1,
      charset=ls.charset
    ),
    dn_str
  )
  if commandbutton:
    return '<table><tr><td>%s</td><td>%s</td></tr></table>' % (
	     CommandButton(
	       form,'read','Read',sid,dn=dn
             ),
	     href_str
	   )
  else:
    return href_str


# Output table containing CommandButton()'s
def CommandTable(
  outf,
  commandlist,
  div_id='CommandDiv',
  table_id='CommandTable',
  table_summary='Command buttons'
):
  if commandlist:
    outf.write("""
    <div id="%s">
    <table
      id="%s"
      summary="%s"
      width="100%%"
      cellpadding="0%%"
      cellspacing="0%%"
    >
    <tr>
    """ % (div_id,table_id,table_summary))
    for s in commandlist:
      outf.write('<td id="MenuItem" align=center valign=top>%s</td>\n' % s)
    outf.write('</tr>\n</table>\n</div>\n')


def ContextMenuSingleEntry(sid,outf,form,ls):
  """Output the context menu for a single entry"""
  CommandTable(outf,[
    CommandButton(
      form,'vcard','vCard',
      sid,
      target='_labeledurl'
    ),
    CommandButton(
      form,'search','Other format',
      sid,
      extrastr="""<br />
      <select name="search_output">
        <option value="ldif">LDIF (Umich style)</option>
        <option value="ldif1">LDIF (Version 1)</option>
	<option value="dsml">DSML</option>
      </select>
      """,
      hidden_fields=[
        ('search_scope','base'),
        ('search_filterstr','(objectclass=*)')
      ],
      target='_altoutput'
    ),
    CommandButton(
      form,'modifyform','Modify',
      sid,
    ),
    CommandButton(
      form,'modrdn','Modify RDN',
      sid,
    ),
    CommandButton(
      form,'delete','Delete',
      sid,
    ),
    CommandButton(
      form,'passwd','Password of entry',
      sid,
      hidden_fields=[
        ('passwd_who',ls.dn),
      ],
    ),
    CommandButton(
      form,'login','Login as',
      sid,
      hidden_fields=[
        ('login_who',ls.dn),
      ],
    ),
  ],div_id='ContextMenuDiv',table_id='ContextMenuTable')


# Output the main menu at top of screen
def MainMenuTable(sid,outf,form,ls,current_dn=''):

  MainMenuList = []

  # convert and store to speed things up
  if ls!=None:

    if current_dn:
      ls.dn = current_dn

    # convert and store to speed things up
    if hasattr(ls,'host') and ls.dn!=None:

      MainMenuList.append(
        CommandButton(
          form,'search','Down',
          sid,
          hidden_fields=[
            ('search_scope','one'),
            ('search_filterstr','(objectclass=*)')
          ],
        )
      )
      if ls.dn:
        # Build the select list for choosing parent DNs
        parentdn_list = ldapbase.explode_dn(ls.dn)[1:]
        parentdn_select_list = []
        for i in xrange(len(parentdn_list)):
          parentdn_select_list.append(
            (
              u','.join(parentdn_list[i:]),
              parentdn_list[i]
            )
          )
        parentdn_select_list.append((u'',u'- World -'))
        parentdn_select_field = cgiforms.formSelectClass(
          'dn',
          u'Parent DN',
          parentdn_select_list,
          default=[parentdn_select_list[0][0]]
        )
        parentdn_select_field.setcharset(form.accept_charset)
        MainMenuList.append(
          CommandButton(
	    form,'search','Up',
            sid,
	    extrastr='<br />%s' % (parentdn_select_field.inputfield()),
            hidden_fields=[
              ('search_scope','one'),
              ('search_filterstr','(objectclass=*)')
            ],
          )
        )
      MainMenuList.append(
        CommandButton(form,'searchform','Search',
	  sid,
	  extrastr="""<br />
          <select name="searchform_mode">
            <option value="base">Base</option>
            <option value="adv">Advanced</option>
            <option value="exp">Expert</option>
          </select>
          """
        )
      )
      MainMenuList.extend(
        [
          CommandButton(form,'read',{0:'Read',1:'RootDSE'}[ls.dn==''],sid),
	  CommandButton(form,'addform','New entry',sid),
        ]
      )
      if hasattr(ls,'who') and ls.who:
        MainMenuList.append(
          CommandButton(
            form,'passwd','Password',sid,
            hidden_fields=[
              ('passwd_who',ls.who),
            ],
          )
        )
      MainMenuList.append(
        CommandButton(form,'conninfo','ConnInfo',sid)
      )
      MainMenuList.append(CommandButton(form,'login','Login',sid))

      # Status line
      if hasattr(ls,'who') and ls.who:
        who_str = ', bound as %s' % DisplayDN(sid,form,ls,ls.who)
      else:
        who_str = ''
      outf.write('<div id="StatusDiv">%s on <strong>%s</strong>: &quot;%s&quot; %s</div>' % (
        DisplayDN(sid,form,ls,ls.dn),
        ls.host.encode(form.accept_charset),
        unicode(w2lcnf.GetParam(ls,'description',''),'iso-8859-1').encode(form.accept_charset),
        who_str,
      ))

  MainMenuList.append(CommandButton(form,'','Disconnect',sid))

  # Main menu
  CommandTable(outf,MainMenuList,div_id='MainMenuDiv',table_id='MainMenuTable')


# Return a pretty HTML-formatted string for describing the attribute
def AttrNameStr(attr):
  if w2lcore.ldap_knownattr.has_key(attr):
    return '<strong>%s</strong><br><em>(%s)</em>' % (
      w2lcore.ldap_knownattr[attr],attr
    )
  else:
    return '<em>%s</em>' % attr


def LDAPURLButton(sid,form,ls,data):
  host,dn,search_attr,search_scope,search_filterstr,ext = ldapbase.parse_ldap_url(data)
  if not host:
    # If the LDAP URL does not contain a host name we stick to current host
    if ls.host:
      host = ls.host
    else:
      host = ''
  if search_scope!=ldap.SCOPE_BASE:
    command_func, command_text, command_hidden_fields = (
      'search','Search',
      [
        ('search_scope',ldapbase.SEARCH_SCOPE_STR[search_scope]),
        ('search_filterstr',search_filterstr)
      ]
    )
  else:
    command_func, command_text, command_hidden_fields = 'read','Read',[]
  return CommandButton(
    form,command_func,command_text,
    sid,
    host=host,
    dn=dn,
    hidden_fields=command_hidden_fields
  )


def MailToURL(mailadr):
  return '<a href="mailto:%s">%s</a>' % (mailadr,mailadr)


def DataStr(sid,form,ls,attrtype,value,valueindex=0,commandbutton=0,oid_reg={}):
  """
  Return a pretty HTML-formatted string of the attribute value
  """
  if not value:
    return '&nbsp;'
  # Try to avoid excessive regex checking by looking
  # at widely used attribute types
  attrtype_lower = attrtype.lower()
  if attrtype_lower in [
    'cn','givenname','sn','c','st','o','ou','l',
    'telephonenumber','homephonenumber','street','description'
  ]:
    return w2lcore.utf2display(form.accept_charset,value)
  elif attrtype_lower=='mail':
    return MailToURL(value)
  elif oid_reg and (attrtype in ROOTDSE_ATTRS_LOWER) and oid_reg.has_key(value):
    name,description,reference = oid_reg[value]
    return '<strong>%(name)s</strong> (%(value)s):<br />%(description)s (see %(reference)s)' % vars()
  elif attrtype_lower in ['jpegphoto','thumbnailphoto','thumbnaillogo']:
    return """
      <img
        src="%s/read?sid=%s&read_attr=%s&read_attrindex=%d"
      >
      <br />
      %d bytes of image data
      """ % (
        form.script_name,sid,
        urllib.quote_plus(attrtype),
        valueindex,len(value),
      )
  elif attrtype_lower in ['audio']:
    mimetype = snd_mimetype(value)
    return """
      <embed
        type="%s"
        autostart="false"
        src="%s/read?sid=%s&read_attr=%s&read_attrindex=%d"
      >
      %d bytes of audio data (%s)
      """ % (
        mimetype,
        form.script_name,sid,
        urllib.quote_plus(attrtype),
        valueindex,
        len(value),
        mimetype
      )
  elif attrtype_lower in w2lcore.ldap_binaryattrkeys:
    return CommandButton(
      form,'read','View/Load',sid,
      hidden_fields=[
        ('read_attr',attrtype),
        ('read_attrindex',str(valueindex))
      ]
    )
  # Check for LDAP-URL
  if ldapbase.is_ldap_url(value):
    if commandbutton:
      commandbuttonstr = LDAPURLButton(sid,form,ls,value)
    else:
      commandbuttonstr = ''
    return '<table><tr><td>%s</td><td><a href="%s" target="_ldapurl">%s</a></td></tr></table>' % (
             commandbuttonstr,
	     value.encode(form.accept_charset),
	     w2lcore.utf2display(form.accept_charset,value)
	   )
  # Check for RFC822 mail address
  rm = mail_regex.match(value)
  if rm!=None:
    return MailToURL(value)
  # Check for DN
  if is_dn(value):
    return DisplayDN(sid,form,ls,value,commandbutton)
  # Check for URL
  rm = url_regex.search(value)
  if rm!=None:
    rm = labeleduri_regex.match(value)
    if rm!=None:
      url,label = value.split(' ',1)
      return '<a href="%s/urlredirect?%s" target="_labeledurl">%s</a>' % (form.script_name,url,w2lcore.utf2display(form.accept_charset,label))
    else:
      return '<a href="%s/urlredirect?%s" target="_labeledurl">%s</a>' % (form.script_name,value,value)
  # Check for timestamp
  rm = timestamp_regex.match(value)
  if rm!=None:
    return utctime.strftimeiso8601(utctime.strptime(value))
  # No special format detected
  return '<br>'.join(
    w2lcore.utf2display(form.accept_charset,value.split('$'))
#    unicode(value,ls.charset).encode(form.accept_charset).split('$')
  )


# Ausdrucken eines HTML-Kopfes mit Titelzeile
def PrintHeader(outf,TitleMsg,HTTP_charset='ISO-8859-1'):
  
  httphelper.SendHeader(
    outf,
    'text/html',
    HTTP_charset,
    expires_offset=web2ldapcnf.misc.sec_expire,
  )

  outf.write("""<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">
<html>
<head>
  <meta name="generator" content="web2ldap %s, see www.web2ldap.de" />
  <title>web2ldap - %s</title>
  <meta http-equiv="Content-Type" content="text/html;charset=%s" />
  <meta name="robots" content="noindex,nofollow" />
  %s
</head>
%s""" % (
    w2lcore.Version,
    TitleMsg,
    HTTP_charset,
    web2ldapcnf.html_head,
    web2ldapcnf.html_bodybegin
  ))
  return


# Ausdrucken eines HTML-Endes
def PrintFooter(f):
  f.write("""<div id="AboutDiv">
  <p align="center">
    <font size="-2">
      Powered by
      <a href="http://www.web2ldap.de/" target="_web2ldap">web2ldap %s</a>
    </font>
  </p>
</div>
</body>
</html>""" % (w2lcore.Version))
  return

BindInput = """
  <fieldset title="User account info">
    <table summary="User account info">
      <tr>
        <td>Bind as</td>
        <td>
          <input name="who" maxlength=200 size=40 value="%s">
        </td>
      </tr>
      <tr>
        <td>with password</td>
        <td>
          <input type="password" name="cred" maxlength=100 size=15 value="">
        </td>
      </tr>
    </table>
  </fieldset>
"""

def SearchRootField(form,ls,name='dn',text=u'Search Root'):
  """Prepare input field for search root"""
  if ls.dn:
    dn_select_list = [ls.dn]+ldapbase.ParentDNList(ls.dn,ls.searchRoot)
  else:
    dn_select_list = []
  dn_select_list = msbase.casepreservingunion(
    dn_select_list,
    filter(lambda dn:len(dn)>0,ls.namingContexts),
    ignorecase=1
  )
  dn_select_list.append((u'',u'- World -'))
  srf = cgiforms.formSelectClass(name,text,dn_select_list,ignorecase=1)
  if hasattr(ls,'searchRoot'):
    srf.setdefault(ls.searchRoot)
  else:
    srf.setdefault(u'')
  srf.setcharset(form.accept_charset)
  return srf


def ExceptionMsg(sid,outf,form,ls,Heading,Msg):
  PrintHeader(outf,'Error',form.accept_charset)
  MainMenuTable(sid,outf,form,ls)
  outf.write("""
<div id="ErrorMessageDiv">
  <h1>%s</h1>
  <p>
    %s
  </p>
  <p>
    For questions or error reports contact:
    <a href="mailto:%s">%s</a>
  </p>
</div>
""" % (Heading,Msg,form.env.get('SERVER_ADMIN',''),form.env.get('SERVER_ADMIN','')))
  PrintFooter(outf)


class ObjectClassSelect(cgiforms.formSelectClass):
  """Select field class for choosing the object class(es)"""
  def __init__(
    self,
    name='ldap_oc',
    text='Object classes',
    default=[],
    required=0,
    accesskey='',
    multiple=1,
    size=12,            # Size of displayed select field
    ignorecase=1,       # Ignore the case when checking input
    multiselect=1,      # Multi select field
  ):
    cgiforms.formSelectClass.__init__(
      self,
      name,
      text,
      w2lcore.ldap_known_objectclasses,
      default,
      required,
      accesskey,
      multiple,
      size,
      ignorecase,
      multiselect,      # Multi select field
      pattern='[\w]+'
    )
  
class DistinguishedNameInput(cgiforms.formInputClass):
  """Input field class for LDAP DNs."""
  def __init__(self,name='dn',text='DN'):
    cgiforms.formInputClass.__init__(
      self,name,text,255,ldapbase.dn_pattern,size=25
    )
