"""
w2llocate.py: Try to locate a LDAP host with various methods.

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>
Distributed under GNU Public License (GPL)
"""

import re,socket,ldap

# Modules shipped with web2ldap
import web2ldapcnf, cgiforms,ldapbase, w2lcore, w2lgui


ldap_hostname_aliases = [
#  'ldap','dsa','x500','ldapdb','nds','openldap'
  'ldap',
]

##############################################################################
# LDAP Service Locator
##############################################################################

LOCATE_NAME_RFC822 = 0
LOCATE_NAME_DCDN   = 1
LOCATE_NAME_DOMAIN = 2

def w2l_Locate(sid,outf,command,form,ls):
  """
  Try to locate a LDAP server in DNS by several heuristics
  """
  # Check if DNS module is present
  try:
    import DNS
  except ImportError:
    raise w2lcore.ErrorExitClass(ls,'Could not import DNS module.')

  DNS.ParseResolvConf()

  locate_name = form.field['locate_name'][0].content.strip()

  # Try to determine the format of the input parameter
  rm = re.match('^%s$' % ldapbase.host_pattern,locate_name,re.I)
  if rm!=None:
    # Use DNS domain directly
    locate_domain = locate_name
    locate_name_type = LOCATE_NAME_DOMAIN
  else:
    rm = re.match(w2lgui.mail_pattern,locate_name,re.I)
    if rm!=None:
      # Use domain part of RFC822 mail address
      locate_domain = locate_name.split('@')[-1]
      locate_name_type = LOCATE_NAME_RFC822
    else:
      rm = ldapbase.dn_regex.match(locate_name)
      if rm!=None:
        # Use dc-style LDAP DN
        locate_domain = ldapbase.dcdn2dnsdomain(locate_name)
        locate_name_type = LOCATE_NAME_DCDN
      else:
        # Parameter not recognized
        raise w2lcore.ErrorExitClass(
          ls,
          'Could not recognize parameter format of &quot;%s&quot;.' % (
            locate_name
          )
        )
    
  dns_list = locate_domain.encode().split('.')

  srv_req = DNS.Request(qtype='srv')

  w2lgui.PrintHeader(outf,'Locate',form.accept_charset)
  w2lgui.MainMenuTable(sid,outf,form,ls)

  outf.write('<div id="MessageDiv">\n')

  for dns_index in range(len(dns_list),1,-1):

    dns_name = '.'.join(dns_list[-dns_index:])
    outf.write('<h1>Domain %s</h1>\n' % (dns_name))

    outf.write('<h2>Search for a SRV RR</h2>\n')
    # Search for a SRV RR of dns_name
    srv_result = srv_req.req('_ldap._tcp.%s' % dns_name)
    # Display SRV search results
    for res in srv_result.answers:
      priority, weight, port, hostname = res['data']
      outf.write('<p>Found %s record: %s:%d (priority %d, weight %d)</p>' % (
          res['typename'],hostname,port,priority,weight,
        )
      )
      try:
        host_address = socket.gethostbyname(dns_name)
      except socket.error:
        outf.write('<p>Did not find IP address for host %s.</p>' % (hostname))
        pass
      else:
        ldap_url = ldapbase.create_ldap_url(
          '%s:%d' % (hostname,port),
          ldapbase.dnsdomain2dcdn(dns_name),
          urlencode=1
        )
        outf.write("""
  <p>IP address found for host name %s: %s</p>
  <table>
    <tr>
      <td>%s</td>
      <td><a href="%s" target="_ldapurl">%s</a></td>
    </tr>
""" % (
            hostname,
            host_address,
            w2lgui.LDAPURLButton(sid,form,ls,ldap_url),
            ldap_url,
            ldap_url,
          )
        )

        if locate_name_type==LOCATE_NAME_RFC822:
          ldap_url = ldapbase.create_ldap_url(
            '%s:%d' % (hostname,port),
            ldapbase.dnsdomain2dcdn(dns_name),
            scope=ldap.SCOPE_SUBTREE,
            filterstr='(mail=%s)' % (locate_name),
            urlencode=1
          )
          outf.write("""    <tr>
        <td>%s</td>
        <td><a href="%s" target="_ldapurl">Search %s</a></td>
      </tr>
  """ % (
              w2lgui.LDAPURLButton(sid,form,ls,ldap_url),
              ldap_url,
              locate_name,
            )
          )
        outf.write('</table>\n')

    outf.write('<h2>Search for well known aliases</h2>\n')
    # Search for well known aliases of LDAP servers under dns_name
    for alias in ldap_hostname_aliases:
      alias_name = '.'.join([alias,dns_name])
      try:
        host_address = socket.gethostbyname(alias_name)
      except socket.error:
        pass
      else:
        ldap_url = 'ldap://%s' % (alias_name)
        outf.write("""
  <p>IP address found for host name %s: %s</p>
  <table>
    <tr>
      <td>%s</td>
      <td><a href="%s" target="_ldapurl">%s</a></td>
    </tr>
  </table>
""" % (
            alias_name,
            host_address,
            w2lgui.LDAPURLButton(sid,form,ls,ldap_url),
            ldap_url,
            ldap_url,
          )
        )

  outf.write('</div>\n')
    
  w2lgui.PrintFooter(outf)
