"""
w2lsearch.py: do a search and return results in several formats

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>
Distributed under GNU Public License (GPL)
"""

import sys,time,types,ldap,ldif,dsml, \
       msbase,cgiforms,httphelper,ldapbase, \
       w2lcore,w2lcnf,web2ldapcnf,w2lgui,w2lsearchform

from msbase import intersection

SizeLimitMsg = """
<p>
  <strong>
    Sizelimit exceeded! Only partial results received.
    Try to refine search.
  </strong>
</p>
"""

def w2l_Search(
  sid,
  outf,
  command,
  form,
  ls,
  search_output = 'table',
  search_scope  = ldap.SCOPE_SUBTREE,
  search_filterstr = None
):
  """
  Search for entries and output results as table, pretty-printable output
  or LDIF formatted
  """

  # Suchparameter sind alle getrennt angegeben
  if 'search_output' in form.inputkeys:
    search_output = form.field['search_output'][0].content
  if 'search_scope' in form.inputkeys:
    search_scope_str = form.field['search_scope'][0].content
    search_scope = ldapbase.SEARCH_SCOPE.get(search_scope_str,ldap.SCOPE_SUBTREE)
  else:
    search_scope_str = ldapbase.SEARCH_SCOPE_STR[search_scope]

  # construct LDAP filter string
  if search_filterstr is None:
    if 'search_filterstr' in form.inputkeys:
      search_filterstr = form.field['search_filterstr'][0].content
    else:
      if 'search_mode' in form.inputkeys:
        search_mode = form.field['search_mode'][0].content
      else:
        search_mode = '(&%s)'
      search_filter = []
      if ('search_option' in form.inputkeys) and \
         ('search_attr' in form.inputkeys) and \
         ('search_string' in form.inputkeys) and \
         (
           len(form.field['search_option'][0].content)== \
	   len(form.field['search_attr'][0].content)== \
	   len(form.field['search_string'][0].content) \
         ):
        for i in range(len(form.field['search_option'][0].content)):
	  if form.field['search_string'][0].content[i]:
	    search_filter.append(
	      form.field['search_option'][0].content[i] % (
    		  form.field['search_attr'][0].content[i].lower(), \
		  ldapbase.escape_filter_chars(form.field['search_string'][0].content[i])
	      )
	    )
      else:
        raise w2lcore.ErrorExitClass(ls,'Invalid or empty search form data.')
      # Teilsuchen werden alle und-verknuepft
      if len(search_filter)==1:
        search_mode = '%s'
      search_filterstr = search_mode % (''.join(search_filter))

  if 'search_resminindex' in form.inputkeys:
    search_resminindex = int(form.field['search_resminindex'][0].content)
  else:
    search_resminindex = 0
  if 'search_resnumber' in form.inputkeys:
    search_resnumber = int(form.field['search_resnumber'][0].content)
  else:
    search_resnumber = 10

  if search_output=='print':
    print_template_filenames_dict = w2lcnf.GetParam(ls,'print_template',None)
    print_cols = w2lcnf.GetParam(ls,'print_cols','4')
    
    if print_template_filenames_dict is None:
      raise w2lcore.ErrorExitClass(ls,'No print templates defined.')

    read_attrs = ['objectclass']
    print_template_str_dict = msbase.CaseinsensitiveStringKeyDict()

    for oc in print_template_filenames_dict.keys():
      try:
	print_template_str_dict[oc] = open(print_template_filenames_dict[oc],'r').read()
      except IOError:
	pass
      else:
        read_attrs = msbase.union(
	  read_attrs,
	  msbase.GrabKeys(print_template_str_dict[oc]).keys
	)

  elif search_output=='table':

    search_tdtemplate = msbase.CaseinsensitiveStringKeyDict(
      w2lcnf.GetParam(ls,'search_tdtemplate',{})
    )
    search_tdtemplate_keys = search_tdtemplate.keys()

    read_attrs = []

    for oc in search_tdtemplate_keys:
      read_attrs = msbase.union(
	read_attrs,
	msbase.GrabKeys(search_tdtemplate[oc]).keys
      )

    search_tablistattrs = w2lcnf.GetParam(ls,'search_tablistattrs',[])
    if search_tablistattrs:
      read_attrs.extend(search_tablistattrs)

    read_attrs.append('objectClass')
    read_attrs.append('displayName')

  elif search_output in ['ldif','ldif1','dsml']:
    # read all attributes
    read_attrs = []

  try:
    # Start asynchronous search call
    ldap_msgid = ls.l.search(
      ls.dn.encode('utf-8'),
      search_scope,
      search_filterstr.encode('utf-8'),
      ldapbase.encode_unicode_list(read_attrs),
      0
    )
  except ldap.FILTER_ERROR:
    # Give the user a chance to edit his bad search filter
    w2lsearchform.w2l_SearchForm(
      sid,outf,command,form,ls,searchform_mode='exp',
      Msg='<p>Error: Bad search filter!</p>',
      search_filterstr=search_filterstr
    )
    return
  except ldap.NO_SUCH_OBJECT,e:
    if ls.dn:
      raise e

  search_ldap_url = ldapbase.create_ldap_url(
    ls.host,
    ls.dn,
    read_attrs,
    search_scope,
    search_filterstr,
    urlencode=1
  )

  if search_output=='table':

    search_maxhits = search_resminindex+search_resnumber
    result_dnlist = []
    mailtolist = []
    SearchWarningMsg = ''

    try:
      ldap_res = ls.l.result(ldap_msgid,0,ls.timeout)
    except ldap.NO_SUCH_OBJECT,e:
      if ls.dn:
        raise e
      else:
        result_data = []
    except ldap.LDAPError,e:
      raise e
    else:
      if not type(ldap_res)==types.TupleType:
        ls.l.abandon(ldap_msgid)
        raise w2lcore.ErrorExitClass(ls,'Wrong result type: "%s"' % repr(ldap_res))
      result_type,result_data = ldap_res
      if not result_type in ['RES_SEARCH_ENTRY','RES_SEARCH_RESULT','RES_SEARCH_REFERENCE']:
        ls.l.abandon(ldap_msgid)
        raise w2lcore.ErrorExitClass(ls,'Wrong result type: "%s"' % (result_type))
      resind = 0
      while result_data and \
            (search_maxhits==0 or (resind<search_maxhits)):
        if result_type=='RES_SEARCH_REFERENCE':
          # received search continuation reference
          result_dnlist.extend(ldapbase.unicode_list(result_data[0][1],'utf-8'))
        elif result_type in ['RES_SEARCH_ENTRY','RES_SEARCH_RESULT']:
          # received search results
          if resind>=search_resminindex:
	    for r in result_data:
              try:
                if type(r)==types.TupleType and len(r)==2:
                  # Search result
                  dn,data=r
	          if data.has_key('mail'):
                    mailtolist.extend(data['mail'])
                  result_dnlist.append((unicode(dn,'utf-8'),data))
                elif type(r)==types.StringType:
                  # Search continuation
                  result_dnlist.append(unicode(r,'utf-8'))
                else:
                  raise ValueError,"LDAP result of invalid type %s." % (type(r))
              except UnicodeError:
                raise w2lcore.ErrorExitClass(
                  ls,
                  'Some data received might been malformed UTF-8 encoded. For security reasons the UTF-8 checking is strict!'
                )
        try:
          result_type,result_data = ls.l.result(ldap_msgid,0,ls.timeout)
        except ldap.PARTIAL_RESULTS,e:
          matched_dn, ldap_urls = ldapbase.extract_referrals(e)
          if matched_dn!=None:
            result_data = [
              (
                matched_dn,
                {
                  'objectClass':['referral'],
                  'ref':ldap_urls
                }
              )
            ]
          break
        except ldap.SIZELIMIT_EXCEEDED,e:
          SearchWarningMsg = SizeLimitMsg
          result_data = []
          break
        resind = resind+1

      if result_data:
        # Only partial results retrieved, undisplayed data available
        ls.l.abandon(ldap_msgid)

    # HACK! If searching the root level returns empty search results
    # the namingContexts is queried
    if ls.dn==u'' and \
       search_scope==ldap.SCOPE_ONELEVEL:
      result_dnlist.extend(map(
        lambda dn: (dn,{}),
        ls.namingContexts
      ))

    result_dnlist.sort()

    # Output search results as table with buttons
    w2lgui.PrintHeader(outf,'Search Results',form.accept_charset)
    w2lgui.MainMenuTable(sid,outf,form,ls)

    if len(result_dnlist):

      ContextMenuList = []

      # There's still data left to be read
      if result_data or search_resminindex:

        prev_resminindex = max(0,search_resminindex-search_resnumber)

        if search_resminindex+search_resnumber<=resind:
          ContextMenuList.append(
	    w2lgui.CommandButton(
	      form,'search','-&gt;&gt;',sid,
              hidden_fields=[
                ('search_filterstr',search_filterstr),
                ('search_resminindex',str(search_resminindex+search_resnumber)),
                ('search_resnumber',str(search_resnumber)),
                ('search_scope',search_scope_str),
              ]
            )
	  )

        if search_resminindex:
          ContextMenuList.append(
	    w2lgui.CommandButton(
	      form,'search','&lt;&lt;-',sid,
              hidden_fields=[
                ('search_filterstr',search_filterstr),
                ('search_resminindex',str(prev_resminindex)),
                ('search_resnumber',str(search_resnumber)),
                ('search_scope',search_scope_str),
              ]
            )
	  )

        ContextMenuList.append(
	  w2lgui.CommandButton(
	    form,'search','Display all',sid,
            hidden_fields=[
              ('search_filterstr',search_filterstr),
              ('search_resnumber','0'),
              ('search_scope',search_scope_str),
            ]
          )
        )

	result_message = """
	  <p>
	    Results %d - %d of %s search with filter &quot;%s&quot;.
	  </p>
	  """ % (
	    search_resminindex+1,
	    resind,
	    search_scope_str.encode(),
	    w2lcore.utf2display(form.accept_charset,search_filterstr)
        )

      else:
	result_message = '<p>%s search with filter &quot;%s&quot; found %d entries.</p>' % (
	  search_scope_str.encode(),
	  w2lcore.utf2display(form.accept_charset,search_filterstr),
	  len(result_dnlist)
	)

      ContextMenuList.extend([
	w2lgui.CommandButton(
	  form,'searchform','Refine Filter',sid,
          hidden_fields=[
            ('searchform_mode','exp'),
            ('search_root',ls.dn),
            ('search_filterstr',search_filterstr),
            ('search_scope',search_scope_str),
          ],
        ),
	w2lgui.CommandButton(
	  form,'search','Other format',sid,
          extrastr="""
	  <select name="search_output">
	    <option value="print" selected>Printable</option>
	    <option value="ldif">LDIF (Umich style)</option>
	    <option value="ldif1">LDIF (Version 1)</option>
	    <option value="dsml">DSML</option>
	  </select>
	  """,
          hidden_fields=[
            ('search_scope',search_scope_str),
            ('search_filterstr',search_filterstr),
          ],
	  target='_altoutput'
	)
      ])

      w2lgui.CommandTable(outf,ContextMenuList,div_id='ContextMenuDiv',table_id='ContextMenuTable')

      outf.write('%s<p>%s</p><A HREF="%s" TARGET="_ldapurl">Search by LDAP URL</A>' % (
          SearchWarningMsg,result_message,
          search_ldap_url.encode()
        )
      )
      if mailtolist:
	outf.write(
          '- <A HREF="mailto:%s?cc=%s">Mail to all Cc:-ed</A>- <A HREF="mailto:?bcc=%s">Mail to all Bcc:-ed</A>' % (
	    mailtolist[0],
	    ','.join(mailtolist[1:]),
	    ','.join(mailtolist)
  	  )
  	)
      outf.write('</p>')

      search_tablistattrs_len = len(search_tablistattrs)

      outf.write('<table id=SearchResultTable summary="Table of search results" WIDTH="100%%" CELLPADDING=0 CELLSPACING=0 BORDER RULES=rows>\n')

      for r in result_dnlist:

        if type(r)==types.UnicodeType:
          # Search continuation
          host,dn,search_attr,search_scope,search_filterstr,ext = ldapbase.parse_ldap_url(r)
          tableentrystr='<td id=SearchResultItem>Referral =&gt; %s</td>' % (
            w2lgui.DataStr(sid,form,ls,'',r,commandbutton=0)
          )

        elif type(r)==types.TupleType and len(r)==2:
          dn,entry = r
          objectclasses = entry.get('objectClass',entry.get('objectclass',[]))
	  tdtemplate_oc = intersection(
	    objectclasses,
	    search_tdtemplate_keys,
            ignorecase=1
	  )

          if entry.has_key('displayName') or entry.has_key('displayname'):
            tableentrystr='<td id=SearchResultItem colspan=%d>%s</td>' % (
              search_tablistattrs_len,
              unicode(
                entry.get(
                  'displayName',
                  entry.get(
                    'displayname',
                    [dn.encode('utf-8')]
                  )
                )[0],
                'utf-8'
              ).encode(form.accept_charset)
            )

          elif tdtemplate_oc:

            template_attrs = []
	    for oc in tdtemplate_oc:
	      template_attrs.extend(msbase.GrabKeys(search_tdtemplate[oc]).keys)
	    tableentry_attrs = intersection(
	      template_attrs,
	      entry.keys(),
	      ignorecase=1
	    )
            if tableentry_attrs:
              # Print entry with the help of pre-defined templates
	      tableentry = msbase.CaseinsensitiveStringKeyDict(default='&nbsp;')
	      for attr in tableentry_attrs:
	        tableentry[attr] = []
	        for value in entry[attr]:
  		  tableentry[attr].append(
                    w2lgui.DataStr(sid,form,ls,attr,value,commandbutton=0)
                  )
	        tableentry[attr] = ', '.join(tableentry[attr])
              tdlist = []
	      for oc in tdtemplate_oc:
	        tdlist.append(search_tdtemplate[oc] % tableentry)
	      tableentrystr='<td id=SearchResultItem colspan=%d>%s</td>\n' % (search_tablistattrs_len,'<br>\n'.join(tdlist))
	    else:
              # Print DN
	      tableentrystr='<td id=SearchResultItem colspan=%d>%s</td>' % (
                search_tablistattrs_len,
                w2lgui.DisplayDN(sid,form,ls,dn,commandbutton=0)
              )

	  elif search_tablistattrs and entry.has_key(search_tablistattrs[0]):
            # Print values of attributes listed in search_tablistattrs
	    tableentry = msbase.CaseinsensitiveStringKeyDict(default='&nbsp;')
	    tableentry_attrs = intersection(
	      search_tablistattrs,
	      entry.keys(),
	      ignorecase=1
	    )
	    for attr in tableentry_attrs:
	      tableentry[attr] = []
	      for value in entry[attr]:
  	        tableentry[attr].append(
                  w2lgui.DataStr(sid,form,ls,attr,value,commandbutton=0)
                )
	      tableentry[attr] = '<br />'.join(tableentry[attr])
	    tdlist = []
	    for attr in search_tablistattrs:
	      tdlist.append(tableentry[attr])
            tableentrystr='<td id=SearchResultItem>%s</td>' % '</td><td id=SearchResultItem>'.join(tdlist)

	  else:
            # Print DN
	    tableentrystr='<td id=SearchResultItem colspan=%d>%s</td>' % (
              search_tablistattrs_len,
              w2lgui.DisplayDN(sid,form,ls,dn,commandbutton=0)
            )
          host,who,cred=None,None,None
        else:
          raise ValueError,"LDAP result of invalid type %s." % (type(r))

	outf.write('<tr><td width="5%%">%s</td><td width="5%%">%s</td>%s</tr>' % (
          w2lgui.CommandButton(
	    form,'search','Down',sid,
	    host=host,dn=dn,
            hidden_fields=[
              ('search_scope','one'),
              ('search_filterstr','(objectclass=*)'),
            ]
          ),
          w2lgui.CommandButton(
	    form,'read','Read',sid,
	    host=host,dn=dn,
	  ),
	  tableentrystr
	)
      )
      outf.write('</table>')

    else:
      w2lgui.CommandTable(outf,[
	w2lgui.CommandButton(
	  form,'searchform','Refine Filter',sid,
          hidden_fields=[
            ('searchform_mode','exp'),
            ('search_filterstr',search_filterstr),
            ('search_scope',search_scope_str),
          ],
        )
      ],div_id='ContextMenuDiv',table_id='ContextMenuTable')
      outf.write("""
        <div id=MessageDiv>
          <p>
            No Entries found with %s search with filter &quot;%s&quot;.
          </p>
        </div>""" % (
          search_scope_str.encode(form.accept_charset),
          w2lcore.utf2display(form.accept_charset,search_filterstr)
        )
      )

    w2lgui.PrintFooter(outf)

  elif search_output=='print':
    result_type,result_dnlist = ls.l.result(ldap_msgid,1)
    if result_type != 'RES_SEARCH_RESULT':
      raise w2lcore.ErrorExitClass(ls,'Wrong result type: "%s"' % (result_type))
    if not result_dnlist:
      w2lsearchform.w2l_SearchForm(
        sid,outf,command,form,ls,searchform_mode='exp',
        Msg='<p>No search results. You might want to modify the search filter/base.</p>',
        search_filterstr=search_filterstr
      )
      return
    result_dnlist.sort()

    table=[]
    for dn,entry in result_dnlist:
      objectclasses = entry.get('objectclass',entry.get('objectClass',[]))
      template_oc = intersection(
        map(lambda s:s.lower(),objectclasses),
	map(lambda s:s.lower(),print_template_str_dict.keys())
      )
      if template_oc:
	tableentry = msbase.DefaultDict('')
	attr_list=entry.keys()
	for attr in attr_list:
	  tableentry[attr] = ', '.join(w2lcore.utf2display(form.accept_charset,entry[attr]))
	table.append(print_template_str_dict[template_oc[0]] % (tableentry))

    # Output search results as pretty-printable table without buttons
    w2lgui.PrintHeader(outf,'Search Results',form.accept_charset)
    outf.write("""<h1>%s</h1>
<table
  id="PrintTable"
  summary="Table with search results formatted for printing"
  width="100%%" cellpadding="3%%" cellspacing=0 border=1 rules="rows">
    <colgroup span="%d">
    """ % (w2lcore.utf2display(form.accept_charset,ls.dn),print_cols))
    tdwidth=100/print_cols
    for i in range(print_cols):
      outf.write('<col width="%d%%">\n' % (tdwidth))
    outf.write('</colgroup>\n')
    for i in range(len(table)):
      if i%print_cols==0:
        outf.write('<tr>')
      outf.write('<td width="%d%%" align="left" valign="top">%s</td>' % (tdwidth,table[i]))
      if i%print_cols==print_cols-1:
        outf.write('</tr>\n')
    if len(table)%print_cols!=print_cols-1:
      outf.write('</tr>\n')

    outf.write('</table>\n')

    w2lgui.PrintFooter(outf)


  elif search_output in ['ldif','ldif1']:

    result_type,result_dnlist = ls.l.result(ldap_msgid,0)
    if result_type !='RES_SEARCH_ENTRY':
      ls.l.abandon(ldap_msgid)
      raise w2lcore.ErrorExitClass(ls,'Wrong result type: "%s"' % (result_type))

    # Output search results as LDIF
    httphelper.SendHeader(outf,'text/plain',charset='US-ASCII')
    if search_output=='ldif1':
      outf.write("""########################################################################
# LDIF export by web2ldap %s, see http://www.web2ldap.de
# Date and time: %s
# Bind-DN: %s
# LDAP-URL of search:
# %s
########################################################################
version: 1

""" % (w2lcore.Version,time.strftime('%A, %Y-%m-%d %H:%M:%S GMT',time.gmtime(time.time()+web2ldapcnf.sec_expire)),ls.who,search_ldap_url))
    while result_dnlist:
      for dn,entry in result_dnlist:
	outf.write('%s\n' % ldif.CreateLDIF(dn,entry,w2lcore.ldap_binaryattrkeys))
      result_type,result_dnlist = ls.l.result(ldap_msgid,0)

  elif search_output=='dsml':

    result_type,result_dnlist = ls.l.result(ldap_msgid,0)
    if result_type !='RES_SEARCH_ENTRY':
      ls.l.abandon(ldap_msgid)
      raise w2lcore.ErrorExitClass(ls,'Wrong result type: "%s"' % (result_type))

    # Output search results as DSML in a rather primitive way.
    # (level 1 producer)
    httphelper.SendHeader(outf,'text/xml',charset='UTF-8')
    outf.write("""<dsml:dsml xmlns:dsml="http://www.dsml.org/DSML">
  <!-- directory entries found by %s -->
  <dsml:directory-entries>
""" % (search_ldap_url))

    while result_dnlist:
      for dn,entry in result_dnlist:
	dsml.CreateDSML(outf,dn,entry,w2lcore.ldap_binaryattrkeys)
      result_type,result_dnlist = ls.l.result(ldap_msgid,0)

    outf.write("""  </dsml:directory-entries>
</dsml:dsml>
""")

